@@ -2,14 +2,14 @@ package vars;
 
 use 5.006;
 
-our $VERSION = '1.01';
+our $VERSION = '1.02';
 
 use warnings::register;
 use strict qw(vars subs);
 
 sub import {
     my $callpack = caller;
-    my ($pack, @imports) = @_;
+    my (undef, @imports) = @_;
     my ($sym, $ch);
     foreach (@imports) {
         if (($ch, $sym) = /^([\$\@\%\*\&])(.+)/) {
