@@ -390,9 +390,11 @@ d_sigaction='undef'
 d_signbit='undef'
 d_sigprocmask='undef'
 d_sigsetjmp='undef'
+d_sin6_scope_id='undef'
 d_sitearch='define'
 d_sitecustomize='undef'
 d_snprintf='undef'
+d_sockaddr_sa_len='undef'
 d_sockatmark='undef'
 d_sockatmarkproto='undef'
 d_socket='define'
